@@ -1,6 +1,6 @@
 package Math::PariBuild;
 
-$VERSION = '2.01080601';
+$VERSION = '2.01080604';
 
 require Exporter;
 @ISA = 'Exporter';
@@ -222,7 +222,7 @@ EOP
 
 EOW
   warn(<<'EOW'), return unless $ENV{MATHPARI_USEFTP};
-FTP session is for debugging only; ignore download.
+FTP session is for debugging only; I'm ignoring the downloaded file.
   (Set $ENV{MATHPARI_USEFTP} to TRUE to actually use the downloaded file.)
 
 EOW
@@ -409,7 +409,7 @@ EOP
       print qq(Downloading `$base_url$file'...\n);
       if ($ftp) {
         $ftp->get($file) or die "Cannot get via FTP (",$ftp->message(),"): $!";
-	$ftp->quit or die "Cannot quit FTP: ", ftp->message();
+	$ftp->quit or warn "Warning: cannot quit FTP: ", $ftp->message();
       } else {
 	my $req = HTTP::Request->new(GET => "$base_url$file");
 	my $resp = $ua->request($req);
@@ -481,6 +481,7 @@ sub patches_for ($) {
 			      'patches/diff_2.1.6_no-common'],
 		 '2.1.7' =>  [
 			($^O =~ /darwin/i ? 'patches/diff_2.1.6_no-common' : ()),
+			($^O eq 'MSWin32' ? 'patches/diff_2.1.7_mingw-w64' : ()),
 			      'patches/patch-pari-unnormalized-float',
 			      'patches/diff_2.1.7_-O',
 			      'patches/diff_2.1.7_div',
@@ -503,22 +504,34 @@ return values of patch commands.
 
 =cut
 
+sub patch_args ($) {
+  return '/' unless $^O =~ /win32/i;
+  my($patch, $p) = (shift, 'utils/inc_h.diff');
+  $p =~ s,/,\\,g;
+  system "$patch -p1 --binary < $p"
+    or warn("... Apparently, your patch takes flag --binary...\n"),
+       return ('\\', '--binary');
+  return '\\';
+}
+
 sub patch_pari {
   my ($dir, $version) = (shift, shift);
   $version = get_pari_version($dir) unless defined $version;
   my @patches = patches_for($version) or return;
   print "Patching...\n";
-  my ($rc, $p) = '';
   my $patch = $Config{gnupatch} || 'patch';
+  my ($dir_sep, @args) = patch_args $patch;
+  my ($rc, $p) = join '; ', $dir_sep, @args, '';
   foreach $p (@patches) {
-    my $cmd = "cd $dir ; $patch -p1 < ../$p";
+    (my $pp = "../$p") =~ s,/,$dir_sep,g;
+    my $cmd = "cd $dir && $patch -p1 @args < $pp";
     print "$cmd\n";
     system "$cmd"
       and warn "...Could not patch: \$?=$?, $!; continuing anyway...\n";
-    $rc .= "'$p' => $?, "
+    $rc .= "'$pp' => $?, "
   }
   print "Finished patching...\n";
-  $rc =~ s/,\s+$//;
+  $rc =~ s/,?\s+$//;
   $rc
 }
 
@@ -721,12 +734,22 @@ sub write_paricfg {
 
 EOP
   my $shellq = ($^O eq 'os2' or $^O =~ /win32/i or $^O eq 'dos') ? q(") : q(');
+  my $datadir = '/usr/local/lib/pari/';
+  { local %ENV;
+    delete $ENV{GP_DATA_DIR};
+    (my $o = `$^X -wle "print shift" "print(default(datadir))" | gp -q` || '')
+       =~ s,/?\n*$,,;
+    $datadir = "$o/" if length $o and -d $o;
+  }
+  # with 2.1.7, it is either in one of $miscdir/galdata $share_prefix/pari/galdata
+  my $miscdir = $datadir;	# Not needed with 2.3.4
+  $datadir .= "galdata/" if $version < 2003000;
   print F <<EOP;
 #define SHELL_Q		'\\$shellq'
 EOP
   print F <<EOP;
-#define GPDATADIR "/usr/local/lib/pari/galdata"
-#define GPMISCDIR "/usr/local/lib/pari"
+#define GPDATADIR "$datadir"
+#define GPMISCDIR "$miscdir"
 
 #define PARI_BYTE_ORDER    $Config{byteorder}
 #define NOEXP2	/* Otherwise elliptic.t:11 rounds differetly, and fails */
@@ -875,6 +898,8 @@ sub find_machine_architecture () {
 	   or $os eq 'freebsd' or $os =~ /^cygwin/) {
     chomp($machine = `uname -m`);
     $machine ||= 'ix86';
+  } elsif (($Config{archname} eq 'MSWin32-x86-multi-thread') && ($Config{cc} =~ /gcc/)) {
+    $machine = 'ix86';
   } elsif (0 and $os =~ /win32/i and not $Config{gccversion}) {
     # Not needed with rename of kernel1.s to kernel1.c?
     $machine = 'port'; # Win32 compilers would not understand the assmebler anyway
@@ -930,7 +955,7 @@ sub find_machine_architecture () {
     }
   } elsif ($os eq 'gnu') {# Cover GNU/Hurd, GNU/kFreeBSD and other GNU userland
     chomp($machine = `uname -m`);
-    $machine = 'ix86' if $machine =~ /^i386-/;
+    $machine = 'ix86' if $machine =~ /^i\d86-/;
   }
 
   $machine = 'port'		# No assembler for 64bit - unless alpha/ia64
@@ -954,7 +979,7 @@ sub find_machine_architecture () {
   } elsif (not defined $machine) {
     chomp($machine = `uname -m`);
   }
-  $machine =~ s/[ix]\d86/ix86/ if defined $machine;
+  $machine =~ s/[ix]\d86(-\w+)?/ix86/ if defined $machine;	# i686-pc
 
   print "...Processor of family `$machine' detected\n";
   return $machine;
@@ -1017,6 +1042,7 @@ sub inline_headers_arr {     # These files are cat()ed to pariinl.h
 	       # ppc is not done yet (2.0.15)
 	($pari_version > 2002007
 		? (ppc		     => ['ppc/asm0.h', 'none/divll.h'],
+                   x86_64            => ['x86_64/asm0.h','none/level1.h'],
 		   ia64		     => ['ia64/asm0.h','ia64/asm1.h'])
 		: ()),
 	       sparcv7	      => ['none/asm0.h','none/level1.h'],
@@ -1113,8 +1139,8 @@ sub choose_and_report_assembler {
 			       ? 'hppa' : 'none'),
 	     );
   my $asmarch = $asmarch{$machine} || $machine; # Temporary only
-  my %skip64 = (alpha => 1, none => 1, ia64 => 1);
-  if (not $skip64{$asmarch}
+  my %skip64 = (alpha => 1, none => 1);
+  if (not ( $skip64{$asmarch} or $asmarch =~ /\D64$/ )
       and ($Config{longsize} || 0) == 8) {
     $asmarch .= '_64';
     $asmarch = 'hppa64' if $asmarch eq 'hppa_64';
@@ -1325,7 +1351,7 @@ sub build_funclists_ourselves ($) {
 		 );
   }
   for my $outfile (keys %recipies) {
-    next if -r $outfile;
+    next if -r "../$outfile";
     my $append = '>';
     for my $step (@{$recipies{$outfile}}) {
       #warn "Running `$^X @$step pari.desc $append ../$outfile-tmp'...\n";
@@ -1589,7 +1615,7 @@ sub code_C_translator {
   my @t = ExtUtils::Constant::constant_types(); # macro defs
   my @tt =
     ExtUtils::Constant::C_constant(
-	  'Math::Pari::func_type', 
+	  'Math::Pari::func_type',
           'func_ord_by_type', undef, undef, undef, undef,
           map {{name => $_, value => $codes{$_}, macro => 1}} keys %codes);
 
